/* Copyright (C) 2009-2025 AO MCST. */

/* В gcc есть ряд builtin'ов, который постфактум используется только для того,
 * чтобы дотащить до как-бы-арх-независимых исходников libunwind информацию,
 * которая по смыслу арх-зависимая. Мы не будем всё это поддерживать
 * в компиляторе, а просто реализуем в виде макросов */

/* sparc
 * gnu'шный номер для регистра %sp */
#define __builtin_dwarf_sp_column() 14

/* sparc
 * Инициализация массива char'ов размерами регистров. Нумерация идёт в gnu'той
 * терминологии (см. файл mdes_sparc_reg.c, таблица mdes_sparc_RegTable) */
#ifdef __LP64__
#define __builtin_init_dwarf_reg_size_table(buff) \
{ \
  int i; \
  char *ptr = (char*)(buff); \
  for (i = 0; i < 32; i++) \
    ptr[i] = 8; \
  for (i = 33; i < 102; i++) \
    ptr[i] = 4; \
}
#else
#define __builtin_init_dwarf_reg_size_table(buff) \
{ \
  int i; \
  char *ptr = (char*)(buff); \
  for (i = 0; i < 102; i++) \
    ptr[i] = 4; \
}
#endif

/* sparc
 * mcstbug #152030.13
 * Для sparc'а все регистры принудительно откачиваются в стек, потому как вся
 * unwind-info описана в терминах того, что регистры от предыдущего окна
 * хранятся в стеке */
#ifdef __LP64__
#define __builtin_unwind_init() \
  asm ("flushw");
#else
#define __builtin_unwind_init() \
  asm ("ta 3");
#endif

/* sparc
 * mcstbug #82489.11 */
#define __builtin_eh_return_data_regno(val) \
  (((val) >= 0 && (val) < 4) ? (val) + 24 : -1)

/* sparc
 * mcstbug #82489.12 */
#ifdef __LP64__
#define __builtin_dwarf_cfa() \
  ({ void *p; \
     asm volatile ("add %%fp, 2047, %0" : "=r"(p)); \
     p; })
#else
#define __builtin_dwarf_cfa() \
  ({ void *p; \
     asm volatile ("mov %%fp, %0" : "=r"(p)); \
     p; })
#endif

/* sparc
 * mcstbug #82489.13 */
#ifdef __LP64__
#define __builtin_frob_return_addr(src) \
  ((void*)(src) - 8)
#else
#define __builtin_frob_return_addr(src) \
  ((void*)(src) - 8)
#endif

/* sparc/e2k
 * mcstbug #82489.14 */
#if defined __sparc__
#define __builtin_eh_return(ival,ptr) \
  ({ \
    asm volatile ("mov %0, %%g1\n\t" \
                  "mov %1, %%g2\n\t" \
                  "mov %%g2, %%i7\n\t" \
                  "restore\n\t" \
                  "retl\n\t" \
                  " add %%sp, %%g1, %%sp" \
                  : : "r" ((long)(ival)), "r" ((void*)(ptr)) : "g1", "g2", "i7"); \
  });
#elif defined __e2k__
#define __builtin_eh_return(ival,ptr) \
  ({ \
    asm volatile ("nop" : : "r" ((long)(ival)), "r" ((void*)(ptr))); \
  });
#endif
